<?php
/**
 * Membership Route
 *
 * @package   rcp
 * @copyright Copyright (c) 2019, Restrict Content Pro team
 * @license   GPL2+
 * @since     1.1
 */

class RCP_REST_API_Membership_Route_V1 extends RCP_REST_API_Route {

	/**
	 * Array of whitelisted membership fields.
	 *
	 * @var array
	 */
	private $membership_fields;

	/**
	 * Get things started
	 *
	 * @since 1.0
	 */
	public function init() {

		$this->id = 'memberships';

		// Whitelist of membership fields that can be altered and their associated query arg name.
		$this->membership_fields = array(
			'ID',
			'customer_id',
			'object_id',
			'object_type',
			'currency',
			'initial_amount',
			'recurring_amount',
			'created_date',
			'trial_end_date',
			'renewed_date',
			'cancellation_date',
			'expiration_date',
			'payment_plan_completed_date',
			'auto_renew',
			'times_billed',
			'maximum_renewals',
			'status',
			'gateway_customer_id',
			'gateway_subscription_id',
			'gateway',
			'signup_method',
			'subscription_key',
			'notes',
			'upgraded_from',
			'disabled'
		);

	}

	/**
	 * @param RCP_Membership $membership
	 *
	 * @since 1.1
	 * @return object
	 */
	private function format_membership( RCP_Membership $membership ) {

		$data = new stdClass();

		$data->id                          = $membership->get_id();
		$data->customer_id                 = $membership->get_customer_id();
		$data->object_id                   = $membership->get_object_id();
		$data->object_type                 = $membership->get_object_type();
		$data->currency                    = $membership->get_currency();
		$data->initial_amount              = $membership->get_initial_amount();
		$data->recurring_amount            = $membership->get_recurring_amount();
		$data->created_date                = $membership->get_created_date( false );
		$data->trial_end_date              = $membership->get_trial_end_date();
		$data->renewed_date                = $membership->get_renewed_date( false );
		$data->cancellation_date           = $membership->get_cancellation_date( false );
		$data->expiration_date             = $membership->get_expiration_date( false );
		$data->payment_plan_completed_date = $membership->get_payment_plan_completed_date();
		$data->auto_renew                  = $membership->is_recurring();
		$data->times_billed                = $membership->get_times_billed();
		$data->maximum_renewals            = $membership->get_maximum_renewals();
		$data->status                      = $membership->get_status();
		$data->gateway_customer_id         = $membership->get_gateway_customer_id();
		$data->gateway_subscription_id     = $membership->get_gateway_subscription_id();
		$data->gateway                     = $membership->get_gateway();
		$data->signup_method               = $membership->get_signup_method();
		$data->subscription_key            = $membership->get_subscription_key();
		$data->notes                       = $membership->get_notes();
		$data->upgraded_from               = $membership->get_upgraded_from();
		$data->disabled                    = $membership->is_disabled();

		return $data;

	}

	/**
	 * Get data
	 *
	 * If the `id` parameter is provided then information about a single membership is retrieved.
	 * Otherwise, an array of membership results is returned.
	 *
	 * @param WP_REST_Request $request
	 *
	 * @since 1.1
	 * @return WP_REST_Response
	 */
	public function get_data( WP_REST_Request $request ) {

		if ( ! empty( $this->query_args['id'] ) ) {

			/**
			 * Get single membership by ID.
			 */

			$membership = rcp_get_membership( absint( $this->query_args['id'] ) );

			if ( empty( $membership ) ) {
				return new WP_REST_Response( new WP_Error( 'invalid_membership', __( 'Invalid membership', 'rcp-rest' ), array( 'status' => 404 ) ), 404 );
			}

			return new WP_REST_Response( $this->format_membership( $membership ) );

		} else {

			/**
			 * Get array of memberships.
			 */
			$memberships = rcp_get_memberships( $this->query_args );

			if ( ! empty( $memberships ) ) {
				$response = array_map( array( $this, 'format_membership' ), $memberships );
			} else {
				$response = new WP_Error( 'no_memberships', __( 'No memberships found', 'rcp-rest-api' ), array( 'status' => 404 ) );
			}

			return new WP_REST_Response( $response );

		}

	}

	/**
	 * Add a new membership
	 *
	 * @param WP_REST_Request $request
	 *
	 * @since 1.1
	 * @return WP_REST_Response ID of the newly created membership on success.
	 */
	public function new_post_data( WP_REST_Request $request ) {

		$args = array();

		foreach ( $this->body as $key => $value ) {

			if ( in_array( $key, $this->membership_fields ) ) {

				$args[ $key ] = wp_slash( $value );

			}

		}

		// Customer ID is required.
		if ( empty( $this->body['customer_id'] ) ) {
			$response = new WP_Error( 'missing_customer_id', __( 'No customer ID supplied', 'rcp-rest' ), array( 'status' => 500 ) );

			return new WP_REST_Response( $response );
		}

		// Object ID is required.
		if ( empty( $this->body['object_id'] ) ) {
			$response = new WP_Error( 'missing_object_id', __( 'No object ID supplied', 'rcp-rest' ), array( 'status' => 500 ) );

			return new WP_REST_Response( $response );
		}

		// Disable other memberships first. @todo maybe remove
		if ( ! rcp_multiple_memberships_enabled() ) {
			//rcp_disable_customer_memberships( absint( $this->body['customer_id'] ) );
		}

		$membership_id = rcp_add_membership( $args );

		if ( empty( $membership_id ) ) {
			return new WP_REST_Response( new WP_Error( 'create_failed', __( 'Failed to add new membership', 'rcp-rest' ), array( 'status' => 500 ) ) );
		}

		return new WP_REST_Response( absint( $membership_id ) );

	}

	/**
	 * Update an existing membership
	 *
	 * @param WP_REST_Request $request
	 *
	 * @since  1.1
	 * @return WP_REST_Response
	 */
	public function update_post_data( WP_REST_Request $request ) {

		if ( isset( $this->body['id'] ) ) {
			$this->body['ID'] = $this->body['id'];
			unset( $this->body['id'] );
		}

		if ( empty( $this->body['ID'] ) ) {
			$response = new WP_Error( 'missing_id', __( 'No membership ID supplied', 'rcp-rest' ), array( 'status' => 500 ) );

			return new WP_REST_Response( $response );
		}

		$updated = rcp_update_membership( absint( $this->body['ID'] ), $this->body );

		if ( $updated ) {
			$response = 1;
		} else {
			$response = new WP_Error( 'update_failed', __( 'Update Failed', 'rcp-rest' ), array( 'status' => 500 ) );
		}

		return new WP_REST_Response( $response );

	}

	/**
	 * Delete a membership
	 *
	 * @param WP_REST_Request $request
	 *
	 * @since  1.1
	 * @return WP_REST_Response
	 */
	public function delete_data( WP_REST_Request $request ) {

		if ( isset( $this->body['id'] ) ) {
			$this->body['ID'] = $this->body['id'];
			unset( $this->body['id'] );
		}

		$membership = rcp_get_membership( absint( $this->body['ID'] ) );

		if ( empty( $membership ) ) {
			return new WP_REST_Response( new WP_Error( 'invalid_membership', __( 'Invalid membership', 'rcp-rest' ), array( 'status' => 404 ) ) );
		}

		$membership->disable();

		return new WP_REST_Response( 1 );

	}

	/**
	 * Determine if authenticated user has permission to access response data
	 *
	 * @since 1.1
	 * @return bool
	 */
	public function can_view() {
		return current_user_can( 'rcp_view_members' );
	}

	/**
	 * Determine if authenticated user has permission to edit data
	 *
	 * @since 1.1
	 * @return bool
	 */
	public function can_edit() {
		return current_user_can( 'rcp_manage_members' );
	}

}