<?php
/**
 * Stripe Payment Gateway
 *
 * @package     Restrict Content Pro
 * @subpackage  Classes/Gateways/Stripe
 * @copyright   Copyright (c) 2017, Pippin Williamson
 * @license     http://opensource.org/licenses/gpl-2.0.php GNU Public License
 * @since       2.1
*/

class RCP_Payment_Gateway_Stripe extends RCP_Payment_Gateway {

	protected $secret_key;
	protected $publishable_key;

	/**
	 * Get things going
	 *
	 * @access public
	 * @since  2.1
	 * @return void
	 */
	public function init() {

		global $rcp_options;

		$this->supports[] = 'one-time';
		$this->supports[] = 'recurring';
		$this->supports[] = 'fees';
		$this->supports[] = 'gateway-submits-form';
		$this->supports[] = 'trial';
		$this->supports[] = 'ajax-payment';

		if( $this->test_mode ) {

			$this->secret_key      = isset( $rcp_options['stripe_test_secret'] )      ? trim( $rcp_options['stripe_test_secret'] )      : '';
			$this->publishable_key = isset( $rcp_options['stripe_test_publishable'] ) ? trim( $rcp_options['stripe_test_publishable'] ) : '';

		} else {

			$this->secret_key      = isset( $rcp_options['stripe_live_secret'] )      ? trim( $rcp_options['stripe_live_secret'] )      : '';
			$this->publishable_key = isset( $rcp_options['stripe_live_publishable'] ) ? trim( $rcp_options['stripe_live_publishable'] ) : '';

		}

		if( ! class_exists( 'Stripe\Stripe' ) ) {
			require_once RCP_PLUGIN_DIR . 'includes/libraries/stripe/init.php';
		}

		\Stripe\Stripe::setApiKey( $this->secret_key );

		\Stripe\Stripe::setApiVersion( '2019-05-16' );

		if ( method_exists( '\Stripe\Stripe', 'setAppInfo' ) ) {
			\Stripe\Stripe::setAppInfo( 'WordPress Restrict Content Pro', RCP_PLUGIN_VERSION, esc_url( site_url() ), 'pp_partner_DxPqC5fdD9vjrf' );
		}
	}

	/**
	 * Create a payment intent or setup intent.
	 *
	 * @since 3.2
	 * @return array|WP_Error
	 */
	public function process_ajax_signup() {

		/**
		 * @var RCP_Payments $rcp_payments_db
		 */
		global $rcp_payments_db;

		$intent          = null;
		$stripe_customer = $this->get_or_create_customer( $this->membership->get_customer_id(), $this->membership->get_customer()->get_user_id() );

		if ( is_wp_error( $stripe_customer ) ) {
			return new WP_Error( $stripe_customer->get_error_code(), sprintf( __( 'Error creating Stripe customer: %s', 'rcp' ), $stripe_customer->get_error_message() ) );
		}

		$this->membership->set_gateway_customer_id( sanitize_text_field( $stripe_customer->id ) );

		$intent_args = array(
			'customer'    => $stripe_customer->id,
			'description' => sprintf( 'RCP: %s', $this->membership->get_membership_level_name() ),
			'metadata'    => array(
				'email'          => $this->email,
				'user_id'        => absint( $this->user_id ),
				'level_id'       => absint( $this->membership->get_object_id() ),
				'level'          => sanitize_text_field( $this->membership->get_membership_level_name() ),
				'key'            => sanitize_text_field( $this->membership->get_subscription_key() ),
				'membership_id'  => absint( $this->membership->get_id() ),
				'rcp_payment_id' => absint( $this->payment->id )
			)
		);

		$intent_options         = array();
		$stripe_connect_user_id = get_option( 'rcp_stripe_connect_account_id', false );

		if ( ! empty( $stripe_connect_user_id ) ) {
			$options['stripe_account'] = $stripe_connect_user_id;
		}

		try {
			if ( ! empty( $this->initial_amount ) ) {
				// Create a payment intent.
				$intent_args = wp_parse_args( $intent_args, array(
					'amount'              => $this->initial_amount * rcp_stripe_get_currency_multiplier(),
					'confirmation_method' => 'automatic',
					'confirm'             => false,
					'currency'            => strtolower( rcp_get_currency() ),
					'setup_future_usage'  => 'off_session'
				) );

				/**
				 * @deprecated 3.2 In favour of `rcp_stripe_create_payment_intent_args`.
				 *
				 * @param array                      $args
				 * @param RCP_Payment_Gateway_Stripe $this
				 */
				$old_charge_args = apply_filters_deprecated( 'rcp_stripe_charge_create_args', array(
					$intent_args,
					$this
				), '3.2', 'rcp_stripe_create_payment_intent_args' );

				// Grab a few compatible arguments from the old charges filter.
				$compatible_keys = array( 'amount', 'currency', 'customer', 'description', 'metadata' );
				foreach ( $compatible_keys as $compatible_key ) {
					$intent_args[ $compatible_key ] = ! empty( $old_charge_args[ $compatible_key ] ) ? $old_charge_args[ $compatible_key ] : $intent_args[ $compatible_key ];
				}

				/**
				 * Filters the payment intent arguments.
				 * This replaces the old `rcp_stripe_charge_create_args` filter.
				 *
				 * @since 3.2
				 */
				$intent_args = apply_filters( 'rcp_stripe_create_payment_intent_args', $intent_args, $this );

				$intent_options['idempotency_key'] = rcp_stripe_generate_idempotency_key( $intent_args );

				$intent = \Stripe\PaymentIntent::create( $intent_args, $intent_options );

			} else {
				// Create a setup intent.
				$intent_args = wp_parse_args( $intent_args, array(
					'usage' => 'off_session'
				) );

				$intent_options['idempotency_key'] = rcp_stripe_generate_idempotency_key( $intent_args );

				$intent = \Stripe\SetupIntent::create( $intent_args, $intent_options );
			}
		} catch ( \Stripe\Error\Base $e ) {
			return $this->get_stripe_error( $e );
		} catch ( Exception $e ) {
			return new WP_Error( $e->getCode(), $e->getMessage() );
		}

		// Store the payment intent ID with the payment.
		$rcp_payments_db->add_meta( $this->payment->id, 'stripe_payment_intent_id', sanitize_text_field( $intent->id ) );

		// Add the client secret to the JSON success data.
		return array(
			'stripe_client_secret' => sanitize_text_field( $intent->client_secret ),
			'stripe_intent_type'   => sanitize_text_field( $intent->object )
		);

	}

	/**
	 * Process registration
	 *
	 * @access public
	 * @since  2.1
	 * @return void
	 */
	public function process_signup() {

		/**
		 * @var RCP_Payments $rcp_payments_db
		 */
		global $rcp_payments_db;

		$payment_intent_id = $rcp_payments_db->get_meta( $this->payment->id, 'stripe_payment_intent_id', true );

		if ( empty( $payment_intent_id ) ) {
			$this->handle_processing_error( new WP_Error( 'missing_stripe_payment_intent', __( 'Missing Stripe payment intent, please try again or contact support if the issue persists.', 'rcp' ) ) );
		}

		try {
			if ( empty( $this->initial_amount ) ) {
				$payment_intent = \Stripe\SetupIntent::retrieve( $payment_intent_id );
			} else {
				$payment_intent = \Stripe\PaymentIntent::retrieve( $payment_intent_id );
			}
		} catch ( Exception $e ) {
			$this->handle_processing_error( $e );
		}

		/**
		 * Set up Stripe customer record and attach it to the membership.
		 */
		$customer = $this->get_or_create_customer();

		/**
		 * Update customer record to set description and metadata.
		 */
		try {
			$description = sprintf( __( 'User ID: %d - User Email: %s - Membership: %s', 'rcp' ), $this->user_id, $this->email, $this->subscription_name );
			if ( strlen( $description ) > 350 ) {
				$description = substr( $description, 0, 350 );
			}

			\Stripe\Customer::update( $customer->id, array(
				'description' => sanitize_text_field( $description ),
				'metadata'    => array(
					'user_id'      => absint( $this->user_id ),
					'email'        => sanitize_text_field( $this->email ),
					'subscription' => sanitize_text_field( $this->subscription_name ),
					'customer_id'  => absint( $this->customer->get_id() )
				)
			) );
		} catch ( Exception $e ) {
			$this->handle_processing_error( $e );
		}

		/**
		 * Initial payment / authorization was successful. Let's get some post-payment stuff done.
		 */

		$member = new RCP_Member( $this->user_id ); // for backwards compatibility only

		/**
		 * Complete the payment record if we have a confirmed payment. This activates the membership.
		 * We also attempt to get the transaction ID from the payment intent charge.
		 *
		 * If the charge isn't actually complete now, the webhook will pick it up later.
		 */
		$payment_data = array(
			'payment_type'   => 'Credit Card',
			'transaction_id' => ''
		);

		$paid = false;

		if ( 'payment_intent' == $payment_intent->object && ! empty( $payment_intent->charges->data[0]['id'] ) && 'succeeded' === $payment_intent->charges->data[0]['status'] ) {
			// Set the transaction ID from the charge.
			$payment_data['transaction_id'] = sanitize_text_field( $payment_intent->charges->data[0]['id'] );
			$payment_data['status']         = 'complete';

			$paid = true;
		} elseif ( 'setup_intent' == $payment_intent->object && empty( $this->initial_amount ) ) {
			// We'll get the transaction ID from the subscription later.
			$payment_data['status'] = 'complete';

			$paid = true;
		} else {
			rcp_log( sprintf( 'Stripe Gateway: payment not immediately verified for intent ID %s - waiting on webhook.' ), $payment_intent->id );
		}

		$rcp_payments_db->update( $this->payment->id, $payment_data );

		/*
		 * We need to refresh our local membership variable because updating the payment record above will have modified it.
		 * This feels hacky AF but I'm doing it for now.
		 */
		$this->membership = rcp_get_membership( $this->membership->get_id() );

		/**
		 * Triggers when a payment is processed by the gateway.
		 *
		 * @param RCP_Member                 $member     Deprecated member object.
		 * @param int                        $payment_id ID of the payment that was just processed.
		 * @param RCP_Payment_Gateway_Stripe $this       Gateway object.
		 */
		do_action( 'rcp_gateway_payment_processed', $member, $this->payment->id, $this );

		/**
		 * Clean up any past due or unpaid subscription.
		 *
		 * We only do this if multiple memberships is not enabled, otherwise we can't be
		 * completely sure which ones we need to keep.
		 */
		if ( ! rcp_multiple_memberships_enabled() ) {
			// Set up array of subscriptions we cancel below so we don't try to cancel the same one twice.
			$cancelled_subscriptions = array();

			// Clean up any past due or unpaid subscriptions. We do this to ensure we don't end up with duplicates.
			$subscriptions = $customer->subscriptions->all( array(
				'expand' => array( 'data.plan.product' )
			) );
			foreach ( $subscriptions->data as $subscription ) {

				// Cancel subscriptions with the RCP metadata present and matching member ID.
				if ( ! empty( $subscription->metadata ) && ! empty( $subscription->metadata['rcp_subscription_level_id'] ) && $this->user_id == $subscription->metadata['rcp_member_id'] ) {
					$subscription->cancel();
					$cancelled_subscriptions[] = $subscription->id;
					rcp_log( sprintf( 'Stripe Gateway: Cancelled Stripe subscription %s.', $subscription->id ) );
					continue;
				}

				/*
				 * This handles subscriptions from before metadata was added. We check the plan name against the
				 * RCP membership level database. If the Stripe plan name matches a sub level name then we cancel it.
				 */
				if ( ! empty( $subscription->plan->product->name ) ) {

					/**
					 * @var RCP_Levels $rcp_levels_db
					 */
					global $rcp_levels_db;

					$level = $rcp_levels_db->get_level_by( 'name', $subscription->plan->product->name );

					// Cancel if this plan name matches an RCP membership level.
					if ( ! empty( $level ) ) {
						$subscription->cancel();
						$cancelled_subscriptions[] = $subscription->id;
						rcp_log( sprintf( 'Stripe Gateway: Cancelled Stripe subscription %s.', $subscription->id ) );
					}

				}
			}
		}

		/**
		 * Attach the payment method to the customer so we can use it again.
		 */
		try {
			$payment_method = \Stripe\PaymentMethod::retrieve( $payment_intent->payment_method );
			$payment_method->attach( array(
				'customer' => $customer->id
			) );

			\Stripe\Customer::update( $customer->id, array(
				'invoice_settings' => array(
					'default_payment_method' => $payment_intent->payment_method,
				)
			) );

			/*
			 * Dedupe payment methods.
			 * If someone re-registers with the same card details they've used in the past, Stripe
			 * will actually create a whole new payment method object with the same fingerprint.
			 * This could result in the same card being added to the customer's payment methods in
			 * Stripe, which is kind of annoying. So we dedupe them to make sure one customer only
			 * has each payment method listed once. Hopefully Stripe will handle this automatically
			 * in the future.
			 *
			 * @link https://github.com/stripe/stripe-payments-demo/issues/45
			 */
			$customer_payment_methods = \Stripe\PaymentMethod::all( array(
				'customer' => $customer->id,
				'type'     => 'card'
			) );
			if ( ! empty( $customer_payment_methods->data ) ) {
				foreach ( $customer_payment_methods->data as $existing_method ) {
					// Detach if the fingerprint matches but payment method ID is different.
					if ( $existing_method->card->fingerprint === $payment_method->card->fingerprint && $existing_method->id != $payment_method->id ) {
						$existing_method->detach();
					}
				}
			}
		} catch ( Exception $e ) {
			$this->handle_processing_error( $e );
		}

		if ( $this->auto_renew ) {

			/**
			 * Set up a recurring subscription in Stripe with a delayed start date.
			 *
			 * All start dates are delayed one cycle because we use a one-time payment for the first charge.
			 */

			// Retrieve or create the plan.
			$plan_id = $this->maybe_create_plan( array(
				'name'           => $this->subscription_name,
				'price'          => $this->amount,
				'interval'       => $this->length_unit,
				'interval_count' => $this->length
			) );

			try {

				// Set up base subscription args.
				if ( ! empty( $this->subscription_start_date ) ) {
					// Use subscription start date if provided. This will be a free trial.
					rcp_log( sprintf( 'Stripe Gateway: Using subscription start date for subscription: %s', $this->subscription_start_date ) );
					$start_date = strtotime( $this->subscription_start_date, current_time( 'timestamp' ) );
				} else {
					// Otherwise, use the calculated expiration date of the membership, modified to current time instead of 23:59.

					$base_date = $this->membership->calculate_expiration( true );
					rcp_log( sprintf( 'Stripe Gateway: Using newly calculated expiration for subscription start date: %s', $base_date ) );

					$timezone = get_option( 'timezone_string' );
					$timezone = ! empty( $timezone ) ? $timezone : 'UTC';
					$datetime = new DateTime( $base_date, new DateTimeZone( $timezone ) );
					$datetime->modify( 'midnight' );
					$current_time = getdate( current_time( 'timestamp' ) );
					$datetime->setTime( $current_time['hours'], $current_time['minutes'], $current_time['seconds'] );
					$start_date = $datetime->getTimestamp();
				}
				$sub_args = array(
					'default_payment_method' => $payment_method->id,
					'plan'                   => $plan_id,
					'prorate'                => false,
					'metadata'               => array(
						'rcp_subscription_level_id' => $this->subscription_id,
						'rcp_member_id'             => $this->user_id,
						'rcp_customer_id'           => $this->customer->get_id(),
						'rcp_membership_id'         => $this->membership->get_id(),
						'rcp_initial_payment_id'    => $this->payment->id
					),
					'trial_end'              => $start_date
				);

				rcp_log( sprintf( 'Stripe Gateway: Creating subscription with %s start date.', $start_date ) );

				$sub_options = array(
					'idempotency_key' => rcp_stripe_generate_idempotency_key( $sub_args )
				);

				$stripe_connect_user_id = get_option( 'rcp_stripe_connect_account_id', false );

				if ( ! empty( $stripe_connect_user_id ) ) {
					$sub_options['stripe_account'] = $stripe_connect_user_id;
				}

				/**
				 * Filters the Stripe subscription arguments.
				 *
				 * @param array                      $sub_args
				 * @param RCP_Payment_Gateway_Stripe $this
				 */
				$sub_args = apply_filters( 'rcp_stripe_create_subscription_args', $sub_args, $this );

				// Create the subscription.
				$subscription = $customer->subscriptions->create( $sub_args, $sub_options );

				// Link the Stripe subscription to the RCP membership.
				$this->membership->set_gateway_subscription_id( $subscription->id );

				// If this all started with a Setup Intent then let's use the subscription ID as the transaction ID.
				if ( 'setup_intent' === $payment_intent->object ) {
					$rcp_payments_db->update( $this->payment->id, array(
						'transaction_id' => sanitize_text_field( $subscription->id )
					) );
				}

			} catch ( Exception $e ) {
				$this->handle_processing_error( $e );
			}

		}

		/**
		 * Triggers after a successful Stripe signup.
		 *
		 * @param int                        $user_id ID of the user account.
		 * @param RCP_Payment_Gateway_Stripe $this    Gateway object.
		 */
		do_action( 'rcp_stripe_signup', $this->user_id, $this );

		// Redirect to the success page.
		wp_redirect( $this->return_url );
		exit;

	}

	/**
	 * Get or create Stripe Customer
	 *
	 * Get the Stripe customer record based on the user registering. This scans their
	 * `gateway_customer_id` fields for a Stripe customer ID and retrieves that if found.
	 *
	 * If no Stripe customer ID is located, a new customer record is created.
	 *
	 * @since 3.2
	 * @return \Stripe\Customer|WP_Error
	 */
	public function get_or_create_customer( $rcp_customer_id = 0, $user_id = 0 ) {

		$customer_exists = false;
		$user_id         = ! empty( $user_id ) ? $user_id : $this->user_id;
		$user            = get_userdata( $user_id );
		$rcp_customer_id = ! empty( $rcp_customer_id ) ? $rcp_customer_id : $this->membership->get_customer()->get_id();
		$customer_id     = rcp_get_customer_gateway_id( $rcp_customer_id, array( 'stripe', 'stripe_checkout' ) );

		if ( $customer_id ) {

			$customer_exists = true;

			try {

				// Update the customer to ensure their card data is up to date
				$customer = \Stripe\Customer::retrieve( $customer_id );

				if ( isset( $customer->deleted ) && $customer->deleted ) {

					// This customer was deleted
					$customer_exists = false;

				}

				// No customer found
			} catch ( Exception $e ) {

				$customer_exists = false;

			}

		}

		if ( empty( $customer_exists ) ) {

			try {

				$customer_args = array(
					'email' => $user->user_email,
					'name'  => sanitize_text_field( trim( $user->first_name . ' ' . $user->last_name ) )
				);

				/**
				 * Filters the customer creation arguments.
				 *
				 * @param array                      $customer_args Arguments used to create the Stripe customer.
				 * @param RCP_Payment_Gateway_Stripe $this          Gatway object.
				 */
				$customer_args = apply_filters( 'rcp_stripe_customer_create_args', $customer_args, $this );

				$customer = \Stripe\Customer::create( $customer_args );

			} catch ( Exception $e ) {

				return new WP_Error( $e->getCode(), $e->getMessage() );

			}

		}

		return $customer;

	}

	/**
	 * Get Stripe error from exception
	 *
	 * This converts the exception into a WP_Error object with a localized error message.
	 *
	 * @param \Stripe\Error\Base $e
	 *
	 * @since 3.2
	 * @return WP_Error
	 */
	protected function get_stripe_error( $e ) {

		$wp_error = new WP_Error();

		if ( method_exists( $e, 'getJsonBody' ) ) {
			$body  = $e->getJsonBody();
			$error = $body['error'];
			$wp_error->add( $error['code'], $this->get_localized_error_message( $error['code'] ) );
		} else {
			$wp_error->add( 'unknown_error', __( 'An unknown error has occurred.', 'rcp' ) );
		}

		return $wp_error;

	}

	/**
	 * Localize common Stripe error messages so they're available for translation.
	 *
	 * @link https://stripe.com/docs/error-codes
	 *
	 * @param string $error_code Stripe error code.
	 *
	 * @since 3.2
	 * @return string
	 */
	protected function get_localized_error_message( $error_code ) {

		switch ( $error_code ) {

			case 'api_key_expired' :
				$message = __( 'Payment gateway connection error.', 'rcp' );
				break;

			case 'card_declined' :
				$message = __( 'The card has been declined.', 'rcp' );
				break;

			case 'email_invalid' :
				$message = __( 'Invalid email address. Please enter a valid email address and try again.', 'rcp' );
				break;

			case 'expired_card' :
				$message = __( 'This card has expired. Please try again with a different payment method.', 'rcp' );
				break;

			case 'incorrect_address' :
				$message = __( 'The supplied billing address is incorrect. Please check the card\'s address or try again with a different card.', 'rcp' );
				break;

			case 'incorrect_cvc' :
				$message = __( 'The card\'s security code is incorrect. Please check the security code or try again with a different card.', 'rcp' );
				break;

			case 'incorrect_number' :
			case 'invalid_number' :
				$message = __( 'The card number is incorrect. Please check the card number or try again with a different card.', 'rcp' );
				break;

			case 'incorrect_zip' :
			case 'postal_code_invalid' :
				$message = __( 'The card\'s postal code is incorrect. Please check the postal code or try again with a different card.', 'rcp' );
				break;

			case 'invalid_cvc' :
				$message = __( 'The card\'s security code is invalid. Please check the security code or try again with a different card.', 'rcp' );
				break;

			case 'invalid_expiry_month' :
				$message = __( 'The card\'s expiration month is incorrect.', 'rcp' );
				break;

			case 'invalid_expiry_year' :
				$message = __( 'The card\'s expiration year is incorrect.', 'rcp' );
				break;

			case 'payment_intent_authentication_failure' :
				$message = __( 'Authentication failure.', 'rcp' );
				break;

			case 'payment_intent_incompatible_payment_method' :
				$message = __( 'This payment method is invalid.', 'rcp' );
				break;

			case 'payment_intent_payment_attempt_failed' :
				$message = __( 'Payment attempt failed.', 'rcp' );
				break;

			case 'setup_intent_authentication_failure' :
				$message = __( 'Setup attempt failed.', 'rcp' );
				break;

			default :
				$message = sprintf( __( 'An error has occurred (code: %s).', 'rcp' ), $error_code );
				break;

		}

		return $message;

	}

	/**
	 * Handle Stripe processing error
	 *
	 * @param \Stripe\Error\Base|Exception|WP_Error $e
	 *
	 * @access protected
	 * @since  2.5
	 * @return void
	 */
	protected function handle_processing_error( $e ) {

		if ( method_exists( $e, 'getJsonBody' ) ) {

			$body                = $e->getJsonBody();
			$err                 = $body['error'];
			$error_code          = ! empty( $err['code'] ) ? $err['code'] : false;
			$this->error_message = $this->get_localized_error_message( $error_code );

		} else {

			$error_code          = is_wp_error( $e ) ? $e->get_error_code() : $e->getCode();
			$this->error_message = is_wp_error( $e ) ? $e->get_error_message() : $e->getMessage();

			// $err is here for backwards compat for the rcp_stripe_signup_payment_failed hook below.
			$err = array(
				'message' => $this->error_message,
				'type'    => 'other',
				'param'   => false,
				'code'    => 'other'
			);

		}

		do_action( 'rcp_registration_failed', $this );
		do_action( 'rcp_stripe_signup_payment_failed', $err, $this );

		$error = '<h4>' . __( 'An error occurred', 'rcp' ) . '</h4>';

		if( ! empty( $error_code ) ) {
			$error .= '<p>' . sprintf( __( 'Error code: %s', 'rcp' ), $error_code ) . '</p>';
		}

		if ( method_exists( $e, 'getHttpStatus' ) ) {
			$error .= "<p>Status: " . $e->getHttpStatus() ."</p>";
		}

		$error .= "<p>Message: " . $this->error_message . "</p>";

		wp_die( $error, __( 'Error', 'rcp' ), array( 'response' => 401 ) );
	}

	/**
	 * Process webhooks
	 *
	 * @access public
	 * @return void
	 */
	public function process_webhooks() {

		if( ! isset( $_GET['listener'] ) || strtolower( $_GET['listener'] ) != 'stripe' ) {
			return;
		}

		rcp_log( 'Starting to process Stripe webhook.' );

		// Ensure listener URL is not cached by W3TC
		if ( ! defined( 'DONOTCACHEPAGE' ) ) {
			define( 'DONOTCACHEPAGE', true );
		}

		// retrieve the request's body and parse it as JSON
		$body          = @file_get_contents( 'php://input' );
		$event_json_id = json_decode( $body );
		$expiration    = '';

		// for extra security, retrieve from the Stripe API
		if ( ! isset( $event_json_id->id ) ) {

			rcp_log( 'Exiting Stripe webhook - no event ID found.' );

			die( 'no event ID found' );

		}

		$rcp_payments = new RCP_Payments();

		$event_id = $event_json_id->id;

		try {

			$event         = \Stripe\Event::retrieve( $event_id );
			$payment_event = $event->data->object;

			rcp_log( sprintf( 'Event ID: %s; Event Type: %s', $event->id, $event->type ) );

			if( empty( $payment_event->customer ) ) {
				rcp_log( 'Exiting Stripe webhook - no customer attached to event.' );

				die( 'no customer attached' );
			}

			$invoice = $customer = $subscription = false;

			if ( ! empty( $payment_event->invoice ) ) {
				$invoice = \Stripe\Invoice::retrieve( $payment_event->invoice );

				if ( ! empty( $invoice->subscription ) ) {
					$subscription = \Stripe\Subscription::retrieve( $invoice->subscription );
				}

			}

			// We can also get the subscription by the object ID in some circumstances.
			if ( empty( $subscription ) && false !== strpos( $payment_event->id, 'sub_' ) ) {
				$subscription = \Stripe\Subscription::retrieve( $payment_event->id );
			}

			// Retrieve the membership by subscription ID.
			if ( ! empty( $subscription ) ) {
				$membership = rcp_get_membership_by( 'gateway_subscription_id', $subscription->id );
			}

			// Retrieve the membership by payment meta (one-time charges only).
			if ( ! empty( $payment_event->metadata->membership_id ) ) {
				$membership = rcp_get_membership( absint( $payment_event->metadata->membership_id ) );
			}

			// Retrieve the membership by customer ID.
			if ( empty( $membership ) && ! rcp_multiple_memberships_enabled() ) {
				$membership = rcp_get_membership_by( 'gateway_customer_id', $payment_event->customer );
			}

			if( empty( $membership ) ) {

				// Grab the customer ID from the old meta keys
				global $wpdb;
				$user_id  = $wpdb->get_var( $wpdb->prepare( "SELECT user_id FROM $wpdb->usermeta WHERE meta_key = '_rcp_stripe_user_id' AND meta_value = %s LIMIT 1", $payment_event->customer ) );
				$customer = rcp_get_customer_by_user_id( $user_id );

				if ( ! empty( $customer ) ) {
					/*
					 * We can only use this function if:
					 * 		- Multiple memberships is disabled; or
					 * 		- The customer only has one membership anyway.
					 */
					if ( ! rcp_multiple_memberships_enabled() || 1 === count( $customer->get_memberships() ) ) {
						$membership = rcp_get_customer_single_membership( $customer->get_id() );
					}
				}

			}

			if( empty( $membership ) ) {
				rcp_log( sprintf( 'Exiting Stripe webhook - membership not found. Customer ID: %s.', $payment_event->customer ), true );

				die( 'no membership ID found' );
			}

			$this->membership = $membership;
			$customer = $membership->get_customer();
			$member   = new RCP_Member( $customer->get_user_id() ); // for backwards compatibility

			rcp_log( sprintf( 'Processing webhook for membership #%d.', $membership->get_id() ) );

			$subscription_level_id = $membership->get_object_id();

			if( ! $subscription_level_id ) {
				rcp_log( 'Exiting Stripe webhook - no membership level ID for membership.', true );

				die( 'no membership level ID for member' );
			}

			if( $event->type == 'customer.subscription.created' ) {
				do_action( 'rcp_webhook_recurring_payment_profile_created', $member, $this );
			}

			if( $event->type == 'charge.succeeded' || $event->type == 'invoice.payment_succeeded' ) {

				rcp_log( sprintf( 'Processing Stripe %s webhook.', $event->type ) );

				// setup payment data
				$payment_data = array(
					'date'                  => date( 'Y-m-d g:i:s', $event->created ),
					'payment_type'          => 'Credit Card',
					'user_id'               => $member->ID,
					'customer_id'           => $customer->get_id(),
					'membership_id'         => $membership->get_id(),
					'amount'                => '',
					'transaction_id'        => '',
					'object_id'             => $subscription_level_id,
					'status'                => 'complete',
					'gateway'               => 'stripe'
				);

				if ( $event->type == 'charge.succeeded' ) {

					// Successful one-time payment
					if ( empty( $payment_event->invoice ) ) {

						$payment_data['amount']         = $payment_event->amount / rcp_stripe_get_currency_multiplier();
						$payment_data['transaction_id'] = $payment_event->id;

					// Successful subscription payment
					} else {

						$payment_data['amount']         = $invoice->amount_due / rcp_stripe_get_currency_multiplier();
						$payment_data['transaction_id'] = $payment_event->id;

						if ( ! empty( $payment_event->discount ) ) {
							$payment_data['discount_code'] = $payment_event->discount->coupon_id;
						}

					}

				}

				if( ! empty( $payment_data['transaction_id'] ) && ! $rcp_payments->payment_exists( $payment_data['transaction_id'] ) ) {

					if ( ! empty( $subscription ) ) {

						$expiration = date( 'Y-m-d 23:59:59', $subscription->current_period_end );
						$membership->set_recurring();
						$membership->set_gateway_subscription_id( $subscription->id );

					}

					$pending_payment_id = rcp_get_membership_meta( $this->membership->get_id(), 'pending_payment_id', true );
					if ( ! empty( $pending_payment_id ) ) {

						// Completing a pending payment. Account activation is handled in rcp_complete_registration()
						$rcp_payments->update( $pending_payment_id, $payment_data );
						$payment_id = $pending_payment_id;

					} else {

						// Inserting a new payment and renewing.
						$membership->renew( $membership->is_recurring(), 'active', $expiration );

						// These must be retrieved after the status is set to active in order for upgrades to work properly
						$payment_data['subscription']     = rcp_get_subscription_name( $membership->get_object_id() );
						$payment_data['subscription_key'] = $membership->get_subscription_key();
						$payment_data['subtotal']         = $payment_data['amount'];
						$payment_data['transaction_type'] = 'renewal';
						$payment_id                       = $rcp_payments->insert( $payment_data );

						if ( $membership->is_recurring() ) {
							do_action( 'rcp_webhook_recurring_payment_processed', $member, $payment_id, $this );
						}

					}

					do_action( 'rcp_gateway_payment_processed', $member, $payment_id, $this );
					do_action( 'rcp_stripe_charge_succeeded', $customer->get_user_id(), $payment_data, $event );

					die( 'rcp_stripe_charge_succeeded action fired successfully' );

				} elseif ( ! empty( $payment_data['transaction_id'] ) && $rcp_payments->payment_exists( $payment_data['transaction_id'] ) ) {

					do_action( 'rcp_ipn_duplicate_payment', $payment_data['transaction_id'], $member, $this );

					die( 'duplicate payment found' );

				}

			}

			// failed payment
			if ( $event->type == 'invoice.payment_failed' ) {

				rcp_log( 'Processing Stripe invoice.payment_failed webhook.' );

				$this->webhook_event_id = $event->id;

				// Make sure this invoice is tied to a subscription and is the user's current subscription.
				if ( ! empty( $event->data->object->subscription ) && $event->data->object->subscription == $membership->get_gateway_subscription_id() ) {
					do_action( 'rcp_recurring_payment_failed', $member, $this );
				} else {
					rcp_log( sprintf( 'Stripe subscription ID %s doesn\'t match membership\'s merchant subscription ID %s. Skipping rcp_recurring_payment_failed hook.', $event->data->object->subscription, $member->get_merchant_subscription_id() ), true );
				}

				do_action( 'rcp_stripe_charge_failed', $payment_event, $event, $member );

				die( 'rcp_stripe_charge_failed action fired successfully' );

			}

			// Cancelled / failed subscription
			if( $event->type == 'customer.subscription.deleted' ) {

				rcp_log( 'Processing Stripe customer.subscription.deleted webhook.' );

				if( $payment_event->id == $membership->get_gateway_subscription_id() ) {

					// If this is a completed payment plan, we can skip any cancellation actions. This is handled in renewals.
					if ( $membership->has_payment_plan() && $membership->at_maximum_renewals() ) {
						rcp_log( sprintf( 'Membership #%d has completed its payment plan - not cancelling.', $membership->get_id() ) );
						die( 'membership payment plan completed' );
					}

					if ( $membership->is_active() ) {
						$membership->cancel();
						$membership->add_note( __( 'Membership cancelled via Stripe webhook.', 'rcp' ) );
					} else {
						rcp_log( sprintf( 'Membership #%d is not active - not cancelling account.', $membership->get_id() ) );
					}

					do_action( 'rcp_webhook_cancel', $member, $this );

					die( 'member cancelled successfully' );

				} else {
					rcp_log( sprintf( 'Payment event ID (%s) doesn\'t match membership\'s merchant subscription ID (%s).', $payment_event->id, $membership->get_gateway_subscription_id() ), true );
				}

			}

			do_action( 'rcp_stripe_' . $event->type, $payment_event, $event );


		} catch ( Exception $e ) {
			// something failed
			rcp_log( sprintf( 'Exiting Stripe webhook due to PHP exception: %s.', $e->getMessage() ), true );

			die( 'PHP exception: ' . $e->getMessage() );
		}

		die( '1' );

	}

	/**
	 * Add credit card fields
	 *
	 * @since 2.1
	 * @return string
	 */
	public function fields() {
		ob_start();
?>

<fieldset class="rcp_card_fieldset">
	<p id="rcp_card_name_wrap">
		<label><?php _e( 'Name on Card', 'rcp' ); ?></label>
		<input type="text" size="20" name="rcp_card_name" class="rcp_card_name card-name" />
	</p>

	<p id="rcp_card_wrap">
		<label><?php esc_html_e( 'Credit Card', 'rcp' ); ?></label>
		<div id="rcp-card-element"></div>
		<div id="rcp-card-element-errors"></div>
	</p>
</fieldset>
<?php
		return ob_get_clean();
	}

	/**
	 * Validate additional fields during registration submission
	 *
	 * @since  2.1
	 * @return void
	 */
	public function validate_fields() {

		global $rcp_options;

		if( empty( $_POST['rcp_card_name'] ) ) {
			rcp_errors()->add( 'missing_card_name', __( 'The card holder name you have entered is invalid', 'rcp' ), 'register' );
		}

		if ( $this->test_mode && ( empty( $rcp_options['stripe_test_secret'] ) || empty( $rcp_options['stripe_test_publishable'] ) ) ) {
			rcp_errors()->add( 'missing_stripe_test_keys', __( 'Missing Stripe test keys. Please enter your test keys to use Stripe in Sandbox Mode.', 'rcp' ), 'register' );
		}

		if ( ! $this->test_mode && ( empty( $rcp_options['stripe_live_secret'] ) || empty( $rcp_options['stripe_live_publishable'] ) ) ) {
			rcp_errors()->add( 'missing_stripe_live_keys', __( 'Missing Stripe live keys. Please enter your live keys to use Stripe in Live Mode.', 'rcp' ), 'register' );
		}

	}

	/**
	 * Load Stripe JS
	 *
	 * @since 2.1
	 * @return void
	 */
	public function scripts() {
		$suffix = ( defined( 'SCRIPT_DEBUG' ) && SCRIPT_DEBUG ) ? '' : '.min';

		// Shared Stripe functionality.
		rcp_stripe_enqueue_scripts(
			array(
				'keys' => array(
					'publishable' => $this->publishable_key,
				),
			)
		);

		// Custom registration form handling.
		// Does not set `rcp-register` as a dependency because those are printed in `wp_footer`
		wp_enqueue_script(
			'rcp-stripe-register',
			RCP_PLUGIN_URL . 'includes/gateways/stripe/js/register' . $suffix . '.js',
			array(
				'jquery',
				'rcp-stripe'
			),
			RCP_PLUGIN_VERSION
		);
	}

	/**
	 * Create plan in Stripe
	 *
	 * @deprecated 3.2 In favour of `maybe_create_plan()`
	 * @see RCP_Payment_Gateway_Stripe::maybe_create_plan()
	 *
	 * @param int $plan_id ID number of the plan.
	 *
	 * @since 2.1
	 * @return bool|string - plan_id if successful, false if not
	 */
	private function create_plan( $plan_id = '' ) {
		global $rcp_options;

		// get all membership level info for this plan
		$plan           = rcp_get_subscription_details( $plan_id );
		$price          = round( $plan->price * rcp_stripe_get_currency_multiplier(), 0 );
		$interval       = $plan->duration_unit;
		$interval_count = $plan->duration;
		$name           = $plan->name;
		$plan_id        = $this->generate_plan_id( $plan );
		$currency       = strtolower( rcp_get_currency() );

		try {

			$product_args = array(
				'name' => $name,
				'type' => 'service'
			);

			$product = \Stripe\Product::create(
				$product_args,
				array(
					'idempotency_key' => rcp_stripe_generate_idempotency_key( $product_args )
				)
			);

			$plan_args = array(
				"amount"         => $price,
				"interval"       => $interval,
				"interval_count" => $interval_count,
				"currency"       => $currency,
				"id"             => $plan_id,
				"product"        => $product->id
			);

			$plan = \Stripe\Plan::create(
				$plan_args,
				array(
					'idempotency_key' => rcp_stripe_generate_idempotency_key( $plan_args )
				)
			);

			// plan successfully created
			return $plan->id;

		} catch ( Exception $e ) {

			$this->handle_processing_error( $e );
		}

	}

	/**
	 * Determine if a plan exists
	 *
	 * @deprecated 3.2 In favour of `maybe_create_plan()`
	 * @see RCP_Payment_Gateway_Stripe::maybe_create_plan()
	 *
	 * @param int $plan The ID number of the plan to check
	 *
	 * @since 2.1
	 * @return bool|string false if the plan doesn't exist, plan id if it does
	 */
	private function plan_exists( $plan ) {

		if ( ! $plan = rcp_get_subscription_details( $plan ) ) {
			return false;
		}

		// fallback to old plan id if the new plan id does not exist
		$old_plan_id = strtolower( str_replace( ' ', '', $plan->name ) );
		$new_plan_id = $this->generate_plan_id( $plan );

		/**
		 * Filters the ID of the plan to check for. If this exists, the new subscription will
		 * use this plan.
		 *
		 * @param string $new_plan_id ID of the Stripe plan to check for.
		 * @param object $plan        Subscription level object.
		 */
		$new_plan_id = apply_filters( 'rcp_stripe_existing_plan_id', $new_plan_id, $plan );

		// check if the plan new plan id structure exists
		try {

			$plan = \Stripe\Plan::retrieve( $new_plan_id );
			return $plan->id;

		} catch ( Exception $e ) {}

		try {
			// fall back to the old plan id structure and verify that the plan metadata also matches
			$stripe_plan = \Stripe\Plan::retrieve( $old_plan_id );

			if ( (int) $stripe_plan->amount !== (int) $plan->price * 100 ) {
				return false;
			}

			if ( $stripe_plan->interval !== $plan->duration_unit ) {
				return false;
			}

			if ( $stripe_plan->interval_count !== intval( $plan->duration ) ) {
				return false;
			}

			return $old_plan_id;

		} catch ( Exception $e ) {
			return false;
		}

	}

	/**
	 * Checks to see if a plan exists with the provided arguments, and if so, returns the ID of
	 * that plan. If not, a new plan is created.
	 *
	 * This method differs from create_plan() and plan_exists() because it doesn't expect
	 * a membership level ID number. This allows for the creation of plans that may not be
	 * exactly based on a membership level's parameters.
	 *
	 * @param array $args           {
	 *                              Array of arguments.
	 *
	 * @type string $name           Required. Name of the plan.
	 * @type float  $price          Required. Price each interval.
	 * @type string $interval       Optional. Billing interval (i.e ."day", "month", "year"). Default is "month".
	 * @type int    $interval_count Optional. Interval count. Default is "1".
	 * @type string $currency       Optional. Currency. Defaults to site currency.
	 * @type string $id             Optional. Plan ID. Automatically generated based on arguments.
	 *                    }
	 *
	 * @since 3.2
	 * @return string|WP_Error Plan ID on success or WP_Error on failure.
	 */
	public function maybe_create_plan( $args ) {

		$args = wp_parse_args( $args, array(
			'name'           => '',
			'price'          => 0.00,
			'interval'       => 'month',
			'interval_count' => 1,
			'currency'       => strtolower( rcp_get_currency() ),
			'id'             => ''
		) );

		// Name and price are required.
		if ( empty( $args['name'] ) || empty( $args['price'] ) ) {
			return new WP_Error( 'missing_name_price', __( 'Missing plan name or price.', 'rcp' ) );
		}

		/*
		 * Create a new object that looks like a membership level object.
		 * We do this because generate_plan_id() expects a membership level object but we
		 * don't actually have one.
		 */
		if ( empty( $args['id'] ) ) {
			$plan_level                = new stdClass();
			$plan_level->name          = $args['name'];
			$plan_level->price         = $args['price'];
			$plan_level->duration      = $args['interval_count'];
			$plan_level->duration_unit = $args['interval'];
			$plan_id                   = $this->generate_plan_id( $plan_level );
		} else {
			$plan_id = $args['id'];
		}

		if ( empty( $plan_id ) ) {
			return new WP_Error( 'empty_plan_id', __( 'Empty plan ID.', 'rcp' ) );
		}

		// Convert price to Stripe format.
		$price = round( $args['price'] * rcp_stripe_get_currency_multiplier(), 0 );

		// First check to see if a plan exists with this ID. If so, return that.

		try {

			$membership_level = isset( $plan_level ) ? $plan_level : new stdClass();

			/**
			 * Filters the ID of the plan to check for. If this exists, the new subscription will
			 * use this plan.
			 *
			 * @param string $plan_id          ID of the Stripe plan to check for.
			 * @param object $membership_level Membership level object.
			 */
			$existing_plan_id = apply_filters( 'rcp_stripe_existing_plan_id', $plan_id, $membership_level );

			$plan = \Stripe\Plan::retrieve( $existing_plan_id );

			return $plan->id;

		} catch ( Exception $e ) {
		}

		// Otherwise, create a new plan.

		try {

			$product = \Stripe\Product::create( array(
				'name' => $args['name'],
				'type' => 'service'
			) );

			$plan = \Stripe\Plan::create( array(
				"amount"         => $price,
				"interval"       => $args['interval'],
				"interval_count" => $args['interval_count'],
				"currency"       => $args['currency'],
				"id"             => $plan_id,
				"product"        => $product->id
			) );

			// plan successfully created
			return $plan->id;

		} catch ( Exception $e ) {

			rcp_log( sprintf( 'Error creating Stripe plan. Code: %s; Message: %s', $e->getCode(), $e->getMessage() ) );

			return new WP_Error( 'stripe_exception', sprintf( 'Error creating Stripe plan. Code: %s; Message: %s', $e->getCode(), $e->getMessage() ) );
		}

	}

	/**
	 * Generate a Stripe plan ID string based on a membership level
	 *
	 * The plan name is set to {levelname}-{price}-{duration}{duration unit}
	 * Strip out invalid characters such as '@', '.', and '()'.
	 * Similar to WP core's sanitize_html_class() & sanitize_key() functions.
	 *
	 * @param object $membership_level
	 *
	 * @since 3.0.3
	 * @return string
	 */
	private function generate_plan_id( $membership_level ) {

		$level_name = strtolower( str_replace( ' ', '', sanitize_title_with_dashes( $membership_level->name ) ) );
		$plan_id    = sprintf( '%s-%s-%s', $level_name, $membership_level->price, $membership_level->duration . $membership_level->duration_unit );
		$plan_id    = preg_replace( '/[^a-z0-9_\-]/', '-', $plan_id );

		return $plan_id;

	}

}
