/* global rcpStripe, rcp_processing, rcp_script_options */

/**
 * Unblock the form, hide loading symbols, and enable registration button.
 */
function rcpStripeEnableForm() {
	jQuery( '#rcp_registration_form #rcp_submit' ).attr( 'disabled', false );
	jQuery( '#rcp_ajax_loading' ).hide();
	jQuery( '#rcp_registration_form' ).unblock();
	jQuery( '#rcp_submit' ).val( rcp_script_options.register );

	rcp_processing = false;
}

/**
 * Handle a failed Stripe payment
 *
 * This does an ajax request to trigger `rcp_registration_failed`, which changes the payment
 * status to `failed` and disables the pending membership.
 *
 * @param {int} payment_id
 * @param {string} message
 *
 * @since 3.2
 */
function rcpStripeHandlePaymentFailure( payment_id, message ) {

	let $ = jQuery;

	$.ajax( {
		type: 'post',
		dataType: 'json',
		url: rcp_script_options.ajaxurl,
		data: {
			action: 'rcp_stripe_handle_initial_payment_failure',
			payment_id: payment_id,
			message: message
		},
		success: function ( response ) { }
	} );

}

/**
 * Close the Stripe Checkout modal.
 *
 * @since 3.2
 */
function rcpStripeCloseCheckoutModal() {
	// Don't allow closing if we're already processing the form.
	if ( rcp_processing ) {
		return;
	}

	let $ = jQuery;
	let modalWrapper = $( '.rcp-modal-wrapper' );
	let modalContainer = $( '.rcp-modal' );

	modalWrapper.fadeOut( 250 );
	modalContainer.hide();

	if ( ! document.getElementById( 'rcp-card-element' ) ) {
		return;
	}

	// Unmount Stripe Elements.
	rcpStripe.elements.card.unmount( '#rcp-card-element' );
}

// Reliant on jQuery triggers, so setup jQuery.
jQuery( function( $ ) {

	/**
	 * Whether or not card details have been entered in Stripe Elements
	 *
	 * @type {boolean}
	 */
	var rcpStripeHasCardDetails = false;

	/**
	 * Prevent registration form from submitting if card details haven't been filled out yet.
	 */
	$( '#rcp_submit' ).click( function ( event ) {
		if ( 'stripe' === rcp_get_gateway().val() && document.getElementById( 'rcp-card-element' ) && ! rcpStripeHasCardDetails ) {
			event.stopPropagation();
			rcpStripeHandleError( rcp_script_options.enter_card_details );
			return false;
		}
	} );

	/**
	 * Attempt to mount the Stripe Elements card when the gateway changes.
	 */
	$( 'body' ).on( 'rcp_gateway_loaded', function( e, gateway ) {
		if ( ! document.getElementById( 'rcp-card-element' ) ) {
			return;
		}

		// Field is available, mount.
		rcpStripe.elements.card.mount( '#rcp-card-element' );

		// Flag as having card details.
		rcpStripe.elements.card.addEventListener( 'change', function ( event ) {
			if ( event.complete ) {
				rcpStripeHasCardDetails = true;
			}
		} );

		// Handle errors.
		rcpStripe.elements.card.addEventListener( 'change', rcpStripeToggleElementErrors );
	} );

	/**
	 * After registration has been processed, handle card payment.
	 */
	$( 'body' ).on( 'rcp_registration_form_processed', function( event, form, response ) {

		// Not on Stripe gateway, bail.
		if ( response.gateway.slug !== 'stripe' ) {
			return;
		}

		/*
		 * Bail if the amount due today is 0 AND:
		 * the recurring amount is 0, or auto renew is off
		 */
		if ( ! response.total && ( ! response.recurring_total || ! response.auto_renew ) ) {
			return;
		}

		// 100% discount, bail.
		if ( $( '.rcp_gateway_fields' ).hasClass( 'rcp_discounted_100' ) ) {
			return;
		}

		// Trigger error if we don't have a client secret.
		if ( ! response.gateway.data.stripe_client_secret ) {
			rcpStripeHandleError( rcp_script_options.error_occurred );
			rcpStripeHandlePaymentFailure( response.payment_id, rcp_script_options.error_occurred );

			return;
		}

		let cardHolderName = $( '.card-name' ).val();

		if ( 'payment_intent' === response.gateway.data.stripe_intent_type ) {
			/**
			 * Payment intent
			 */
			rcpStripe.Stripe.handleCardPayment(
				response.gateway.data.stripe_client_secret, rcpStripe.elements.card, {
					payment_method_data: {
						billing_details: { name: cardHolderName }
					}
				}
			).then( function( paymentResult ) {
				if ( paymentResult.error ) {
					rcpStripeHandleError( paymentResult.error.message );
					rcpStripeHandlePaymentFailure( response.payment_id, paymentResult.error.message );
				} else {
					rcp_submit_registration_form( form, response );
				}
			} );
		} else {
			/**
			 * Create a setup intent if the payment due today is $0 but we're creating a subscription.
			 */
			rcpStripe.Stripe.handleCardSetup(
				response.gateway.data.stripe_client_secret, rcpStripe.elements.card, {
					payment_method_data: {
						billing_details: { name: cardHolderName }
					}
				}
			).then( function( setupResult ) {
				if ( setupResult.error ) {
					rcpStripeHandleError( setupResult.error.message );
					rcpStripeHandlePaymentFailure( response.payment_id, setupResult.error.message );
				} else {
					// Success
					rcp_submit_registration_form( form, response );
				}
			} );
		}
	} );

	/**
	 * [register_form_stripe] Stripe Checkout modal
	 */
	$( 'body' ).on( 'click', '.rcp-stripe-register-submit-button', function ( e ) {
		e.preventDefault();

		let modalWrapper = $( '.rcp-modal-wrapper' );
		let modalContainer = $( '.rcp-modal' );
		let form = $( this ).parents( '.rcp-stripe-register' );

		// Mount the Stripe Elements card.
		if ( ! document.getElementById( 'rcp-card-element' ) ) {
			return;
		}

		// Fade in wrapper.
		modalWrapper.fadeIn( 250 );

		// Field is available, mount.
		rcpStripe.elements.card.mount( '#rcp-card-element' );

		let registerButton = modalContainer.find( 'input.rcp-modal-submit' );

		// Flag as having card details.
		rcpStripe.elements.card.addEventListener( 'change', function ( event ) {
			if ( event.complete ) {
				rcpStripeHasCardDetails = true;
			}
		} );

		// Handle errors.
		rcpStripe.elements.card.addEventListener( 'change', rcpStripeToggleElementErrors );

		// Set form content.
		if ( form.data( 'name' ) ) {
			$( '.rcp-modal-membership-name' ).text( form.data( 'name' ) ).show();
		}
		if ( form.data( 'description' ) ) {
			$( '.rcp-modal-membership-description' ).text( form.data( 'description' ) ).show();
		}
		if ( form.data( 'panel-label' ) ) {
			registerButton.val( form.data( 'panel-label' ) );
		}
		if ( form.data( 'level-id' ) ) {
			modalContainer.find( '#rcp-stripe-checkout-level-id' ).val( form.data( 'level-id' ) );
		}

		// Fade in container.
		modalContainer.fadeIn( 250 );

		/**
		 * Close the modal on these three events...
		 */

		// If they click the close icon, close modal.
		modalWrapper.find( '.rcp-modal-close' ).on( 'click', function() {
			rcpStripeCloseCheckoutModal();
		} );

		// If they press "ESC" on the keyboard, close modal.
		$( document ).on( 'keydown', function ( e ) {
			if ( 27 === e.keyCode ) {
				rcpStripeCloseCheckoutModal();
			}
		} );

		// If they click outside the container, close modal.
		$( document ).mouseup( function( e ) {
			if ( ! modalContainer.is( e.target ) && modalContainer.has( e.target ).length === 0 ) {
				rcpStripeCloseCheckoutModal();
			}
		} );
	} );
} );
